<?php

namespace App\Tests\functional;

use App\Entity\Capsule;
use App\Entity\User;
use App\Repository\UserRepository;
use Exception;
use Symfony\Bundle\FrameworkBundle\KernelBrowser;
use Symfony\Bundle\FrameworkBundle\Test\WebTestCase;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Component\Mime\RawMessage;

class RegistrationControllerTest extends WebTestCase
{
    private KernelBrowser $client;

    protected function setUp(): void
    {
        self::ensureKernelShutdown();
        $this->client = static::createClient();
    }

    public function testNewUserRegistrationShouldAcceptGeneralConditions(): void
    {
        $userEmail =  'newUser@localhost.com';

        $this->registerUser($userEmail, $this->client);

        $user = $this->getUserByEmail($userEmail);

        $this->assertEquals(
            true,
            $user->hasAcceptGeneralConditions(),
            'The user should accept the general conditions before registration'
        );
    }

    public function testNewUserRegistrationShouldBeNotifiedOfAccountValidationByMail(): void
    {
        $userEmail =  'newUser@localhost.com';

        $this->registerUser($userEmail, $this->client);

        $this->assertUserIsRedirectedToTheEmailNotificationPage($this->client);
    }

    public function testNewUserRegistrationShouldHaveUnverifiedAccount(): void
    {
        $userEmail =  'newUser@localhost.com';

        $this->registerUser($userEmail, $this->client);

        $user = $this->getUserByEmail($userEmail);

        $this->assertFalse(
            $user->IsVerified(),
            'The user should have his account disabled, until he validates his email.'
        );
    }

    public function testWhenNewUserRegistersThenTheSystemShouldSendAnEmailToTheUser(): void
    {
        $userEmail =  'newUser@localhost.com';

        $this->registerUser($userEmail, $this->client);

        $this->checkEmailHasBeenSentAndGetEmailMessage($userEmail);
    }

    public function testEmailValidationRegistrationShouldEnableUser(): void
    {
        $userEmail =  'newUser@localhost.com';

        $this->registerUser($userEmail, $this->client);
        $emailMessage = $this->checkEmailHasBeenSentAndGetEmailMessage($userEmail);
        $this->client->followRedirect();

        $this->clickOnEmailMessageLink($emailMessage, $this->client);

        // Check the user has been redirected on login page
        $this->assertResponseRedirects(
            '/login',
            302,
            'Once the user email is validated, memorekall should redirect the user to the login page'
        );

        // Check the user is now able to login
        $user = $this->getUserByEmail($userEmail);
        $this->assertTrue(
            $user->isVerified(),
            'Once the user email is validated, he should be able to login'
        );

        // Check the redirection is successful
        $this->client->followRedirect();
        $this->assertResponseIsSuccessful('Once the user has validated his email, he should be on the login page');
    }

    public function testAnAuthenticatedUserShouldNotBeAbleToAccessTheRegisterPage(): void
    {
        $testUser = $this->getUserByEmail('defaultUser@localhost.com');

        $this->client->loginUser($testUser);
        $this->client->request('GET', '/register');

        $this->assertResponseRedirects('/en/my_capsules', 302);
    }

    public function testAnAuthenticatedUserShouldNotBeAbleToAccessTheLoginPage(): void
    {
        $testUser = $this->getUserByEmail('defaultUser@localhost.com');

        $this->client->loginUser($testUser);
        $this->client->request('GET', '/login');

        $this->assertResponseRedirects('/en', 302);
    }

    public function testSubmittingTheRegisterFormWithAnInvalidEmailAddressShouldDisplayAFeedbackError(): void
    {
        $userEmail =  'invalidEmailAddress';

        $crawler = $this->registerUser($userEmail, $this->client);
        $html = $crawler->outerHtml();

        $this->assertStringContainsString("is not a valid email", $html);
    }

    /**
     * @param string $userEmail The registered user email
     * @return RawMessage The email message sent to the user
     * @throws Exception
     */
    private function checkEmailHasBeenSentAndGetEmailMessage(string $userEmail): RawMessage
    {
        $this->assertEmailCount(
            1,
            null,
            'Once the user has been registered, the system should send an email to this user'
        );
        $emailMessage = $this->getMailerMessage(0);

        if (null === $emailMessage) {
            throw new Exception("Email message could not be found");
        }

        $this->assertEmailAddressContains(
            $emailMessage,
            'To',
            $userEmail,
            'Once the user has been registered, the system should send an email to this user'
        );

        return $emailMessage;
    }

    private function assertUserIsRedirectedToTheEmailNotificationPage(KernelBrowser $client): void
    {
        $this->assertResponseRedirects(
            '/register_mail_sent',
            302,
            'Once the user has been registered,' .
            ' he should be redirected to a webpage noticing' .
            ' an email has been sent for account validation'
        );
        $client->followRedirect();
        $this->assertResponseIsSuccessful('/register_mail_sent');
    }

    private function registerUser(string $userEmail, KernelBrowser &$client): Crawler
    {
        $crawler = $client->request('GET', '/register');

        $this->assertResponseIsSuccessful();

        $client->enableProfiler();

        $submit = $crawler->selectButton('Register');
        $form = $submit->form();

        $form['registration_form[firstName]'] = 'new User FirstName';
        $form['registration_form[lastName]'] = 'new User LastName';
        $form['registration_form[email]'] = $userEmail;
        $form['registration_form[plainPassword][first]'] = 'password';
        $form['registration_form[plainPassword][second]'] = 'password';
        $form['registration_form[acceptGeneralConditions]'] = "1";

        return $client->submit($form);
    }

    private function clickOnEmailMessageLink(RawMessage $emailMessage, KernelBrowser $client): void
    {
        /** @phpstan-ignore-next-line */
        $crawler = new Crawler($emailMessage->getHtmlBody());
        $crawlerLink = $crawler->selectLink('Confirm my email')->link();
        $client->click($crawlerLink);
    }

    private function getUserByEmail(string $userEmail): User
    {
        $userRepository = static::getContainer()->get(UserRepository::class);
        $user = $userRepository->findOneByEmail($userEmail);

        if (! $user instanceof User) {
            throw new Exception("User does not exist.");
        }

        return $user;
    }
}
